import os
import warnings

import click
import pandas as pd
import xlsxwriter

import xlsxreport
from xlsxreport.excel_writer import TableSectionWriter
from xlsxreport.template import ReportTemplate
from xlsxreport.compiler import prepare_table_sections


OUTFILE_DESCRIPTION = (
    "Name of the report file, by default the `INFILE` name is used with the file "
    "extension replaced by '.report.xlsx'."
)
OUTPATH_DESCRIPTION = (
    "Output path of the report file. If specified overrides the `outfile` option."
)
SEPARATOR_DESCRIPTION = "Delimiter to use for the input file, default is \\t."


@click.command()
@click.argument("infile", type=click.Path(exists=True))
@click.argument("config")
@click.option("--outfile", default="", help=OUTFILE_DESCRIPTION)
@click.option("--outpath", default="", help=OUTPATH_DESCRIPTION)
@click.option("--sep", default="\t", help=SEPARATOR_DESCRIPTION)
def cli(infile: str, config: str, outfile: str, outpath: str, sep: str) -> None:
    """Create a formatted Excel report from csv INFILE and a formatting CONFIG file.

    The CONFIG argument is first used to look for a file with the specified filepath. If
    no file is found, the XlsxReport appdata directory is searched for a file with the
    corresponding name.
    """
    if os.path.isfile(config):
        config_path = config
    elif xlsxreport.get_config_file(config) is not None:
        config_path = xlsxreport.get_config_file(config)
    else:
        raise click.ClickException(f"Invalid value for `CONFIG`: '{config}' not found.")

    if outpath:
        if not os.path.isdir(os.path.dirname(outpath)):
            outdir = os.path.dirname(outpath)
            raise click.ClickException(
                f"Invalid value for `outpath`: '{outdir}' directory not found."
            )
        report_path = outpath
    elif outfile:
        report_path = os.path.join(os.path.dirname(infile), outfile)
    else:
        infilename = os.path.basename(infile)
        outfilename = ".".join(infilename.split(".")[:-1]) + ".report.xlsx"
        report_path = os.path.join(os.path.dirname(infile), outfilename)

    click.echo(f"Generating formatted Excel report:")
    click.echo(f"----------------------------------")
    click.echo(f"Input file:  {infile}")
    click.echo(f"Config file: {config_path}")

    with warnings.catch_warnings():
        warnings.simplefilter(action="ignore", category=pd.errors.DtypeWarning)
        table = pd.read_csv(infile, sep=sep)

    report_template = ReportTemplate.load(config_path)
    report_template.settings["evaluate_log2_transformation"] = True
    table_sections = prepare_table_sections(report_template, table)

    with xlsxwriter.Workbook(report_path) as workbook:
        worksheet = workbook.add_worksheet("Proteins")
        section_writer = TableSectionWriter(workbook)
        section_writer.write_sections(
            worksheet, table_sections, settings=report_template.settings
        )
    click.echo(f"Report file: {report_path}")


if __name__ == "__main__":
    cli()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
XlsxReport is a Python package to automatically generate formatted excel reports
from quantitative mass spectrometry result tables. YAML config files are used to
describe the content of a result table and the format of the excel report.
"""


from .writer import Reportbook
from .appdir import locate_data_dir, get_config_file, setup_data_dir


__author__ = "David M. Hollenstein"
__license__ = "Apache 2.0"
__version__ = "0.0.8"



import numpy as np
import pytest
import pandas as pd

import xlsxreport.compiler as compiler
from xlsxreport.template import ReportTemplate


@pytest.fixture()
def report_template() -> ReportTemplate:
    standard_section_template = {
        "format": "str",
        "columns": ["Column 1", "Column 2", "Column 3"],
        "column_format": {"Column 1": "float"},
        "column_conditional": {"Column 1": "cond_1"},
        "supheader": "Supheader",
        "conditional": "cond_2",
    }

    tag_sample_section_template = {
        "format": "float",
        "tag": "Tag",
        "supheader": "Supheader",
        "conditional": "cond_1",
        "remove_tag": True,
        "log2": True,
    }

    report_template = ReportTemplate(
        sections={
            "Standard section 1": standard_section_template,
            "Tag sample section 1": tag_sample_section_template,
        },
        formats={
            "header": {"bold": True, "align": "center"},
            "supheader": {"bold": True, "align": "center", "text_wrap": True},
            "str": {"align": "left", "num_format": "@"},
            "float": {"align": "center", "num_format": "0.00"},
        },
        conditional_formats={
            "cond_1": {"type": "2_color_scale"},
            "cond_2": {"type": "3_color_scale"},
        },
        settings={
            "column_width": 10,
            "sample_extraction_tag": "Tag",
            "log2_tag": "[log2]",
            "remove_duplicate_columns": True,
        },
    )
    return report_template


@pytest.fixture()
def example_table() -> pd.DataFrame:
    example_table = pd.DataFrame(
        {
            "Column 1": [1, 2, 3],
            "Column 2": ["A", "B", "C"],
            "Column 4": ["A", "B", "C"],
            "Tag Sample 1": [1, 2, 3],
            "Tag Sample 2": [1, 2, 3],
        }
    )
    return example_table


@pytest.fixture()
def standard_table_section(report_template, example_table) -> compiler.TableSection:
    table_section = compiler.TableSection(
        data=example_table[["Column 1", "Column 2"]].copy(),
        column_formats={
            "Column 1": report_template.formats["float"],
            "Column 2": report_template.formats["str"],
        },
        column_conditionals={
            "Column 1": report_template.conditional_formats["cond_1"],
            "Column 2": {},
        },
        column_widths={
            "Column 1": 10,
            "Column 2": 10,
        },
        headers={"Column 1": "Column 1", "Column 2": "Column 2"},
        header_formats={
            "Column 1": report_template.formats["header"],
            "Column 2": report_template.formats["header"],
        },
        supheader=report_template.sections["Standard section 1"]["supheader"],
        supheader_format=report_template.formats["supheader"],
        section_conditional=report_template.conditional_formats["cond_2"],
    )
    return table_section


@pytest.fixture()
def tag_sample_table_section(report_template, example_table) -> compiler.TableSection:
    table_section = compiler.TableSection(
        data=np.log2(example_table[["Tag Sample 1", "Tag Sample 2"]].copy()),
        column_formats={
            "Tag Sample 1": report_template.formats["float"],
            "Tag Sample 2": report_template.formats["float"],
        },
        column_conditionals={
            "Tag Sample 1": {},
            "Tag Sample 2": {},
        },
        column_widths={
            "Tag Sample 1": 10,
            "Tag Sample 2": 10,
        },
        header_formats={
            "Tag Sample 1": report_template.formats["header"],
            "Tag Sample 2": report_template.formats["header"],
        },
        headers={"Tag Sample 1": "Sample 1", "Tag Sample 2": "Sample 2"},
        supheader="Supheader [log2]",
        supheader_format=report_template.formats["supheader"],
        section_conditional=report_template.conditional_formats["cond_1"],
    )
    return table_section


class TestEvalData:
    def test_data_frame_contains_only_selected_columns(self):
        table = pd.DataFrame({"Column 1": [1, 2, None], "Column 2": ["A", "B", "C"]})
        evaluated_data = compiler.eval_data(table, ["Column 1"])
        assert list(evaluated_data.columns) == ["Column 1"]

    def test_nan_values_in_dataframe_replaced(self):
        table = pd.DataFrame({"Column 1": [1, 2, None], "Column 2": ["A", "B", "C"]})
        evaluated_data = compiler.eval_data(table, ["Column 1"])
        assert not evaluated_data.isna().values.any()
        assert evaluated_data["Column 1"][2] == compiler.NAN_REPLACEMENT_SYMBOL


class TestEvalDataWithLog2Transformation:
    def test_log2_transformation_applied_when_specified(self):
        table = pd.DataFrame({"Column 1": [1, 2, 3], "Column 2": ["A", "B", "C"]})
        evaluated_data = compiler.eval_data_with_log2_transformation(
            table, ["Column 1"], {"log2": True}, evaluate_log_state=False
        )
        assert evaluated_data["Column 1"].tolist() == np.log2(table["Column 1"]).tolist()  # fmt: skip

    def test_log2_transformation_replaces_values_smaller_or_equal_to_zero_with_nan(self):  # fmt: skip
        table = pd.DataFrame({"Column 1": [-1, 0, 1], "Column 2": ["A", "B", "C"]})
        evaluated_data = compiler.eval_data_with_log2_transformation(
            table, ["Column 1"], {"log2": True}, evaluate_log_state=False
        )
        assert evaluated_data["Column 1"].tolist() == ["", "", 0]

    def test_log_space_evaluation_prevents_log2_transformation_of_small_values(self):  # fmt: skip
        table = pd.DataFrame({"Column 1": [1, 2, 3], "Column 2": ["A", "B", "C"]})
        evaluated_data = compiler.eval_data_with_log2_transformation(
            table, ["Column 1"], {"log2": True}, evaluate_log_state=True
        )
        assert evaluated_data["Column 1"].tolist() == [1, 2, 3]

    def test_log_space_evaluation_does_not_prevent_log2_transformation_of_large_values(self):  # fmt: skip
        table = pd.DataFrame({"Column 1": [1, 2, 65], "Column 2": ["A", "B", "C"]})
        evaluated_data = compiler.eval_data_with_log2_transformation(
            table, ["Column 1"], {"log2": True}, evaluate_log_state=True
        )
        assert evaluated_data["Column 1"].tolist() == np.log2(table["Column 1"]).tolist()  # fmt:skip


def test_eval_standard_section_columns_selects_correct_columns():
    template_section = {"columns": ["Column 1", "Column 2", "Column 3"]}
    columns = ["Column 1", "Column 2", "Column 4"]
    selected_columns = compiler.eval_standard_section_columns(columns, template_section)
    assert selected_columns == ["Column 1", "Column 2"]


def test_eval_tag_sample_section_columns_selects_correct_columns():
    template_section = {"tag": "Tag"}
    columns = ["Tag Sample 1", "Tag Sample 2", "Atag Sample 1", "Atag Sample 2", "Atag Sample"]  # fmt: skip
    extraction_tag = "Atag"
    selected_columns = compiler.eval_tag_sample_section_columns(
        columns, template_section, extraction_tag
    )
    assert selected_columns == ["Tag Sample 1", "Tag Sample 2"]


def test_eval_comparison_groups_extracts_correct_values():
    template_section = {
        "comparison_group": True,
        "tag": " vs ",
        "columns": ["P", "A"],
    }
    columns = ["P", "A", "P ex1 vs ex2", "A ex1 vs ex2", "P ex1 vs EX3", "A ex1 vs EX3"]
    comparison_groups = compiler.eval_comparison_groups(columns, template_section)
    assert comparison_groups == ["ex1 vs ex2", "ex1 vs EX3"]


class TestEvalComparisonGroupColumns:
    def test_correct_columns_are_selected(self):
        template_section = {
            "comparison_group": True,
            "tag": " vs ",
            "columns": ["P", "A"],
        }
        columns = ["P", "A", "P ex1 vs ex2", "A ex1 vs ex2", "P ex1 vs EX3", "A ex1 vs EX3"]  # fmt: skip
        selected_columns = compiler.eval_comparison_group_columns(columns, template_section, "ex1 vs EX3")  # fmt: skip
        expected_columns = ["P ex1 vs EX3", "A ex1 vs EX3"]
        assert selected_columns == expected_columns

    def test_column_order_of_section_template_is_used(self):
        template_section = {
            "comparison_group": True,
            "tag": " vs ",
            "columns": ["P", "A"],
        }
        columns = ["A ex1 vs ex2", "P ex1 vs ex2"]  # fmt: skip
        selected_columns = compiler.eval_comparison_group_columns(columns, template_section, "ex1 vs ex2")  # fmt: skip
        expected_columns = ["P ex1 vs ex2", "A ex1 vs ex2"]
        assert selected_columns == expected_columns


class TestEvalComparisonGroupHeaders:
    @pytest.fixture(autouse=True)
    def _init_inputs(self):
        self.columns = ["P ex1 vs ex2", "A ex1 vs ex2", "C ex1 vs ex2"]
        self.comparison_group = "ex1 vs ex2"
        self.section_template = {"tag": " vs "}

    def test_by_default_columns_are_headers(self):
        headers = compiler.eval_comparison_group_headers(
            self.columns, self.section_template, self.comparison_group
        )
        assert headers == {c: c for c in self.columns}

    def test_that_remove_tag_removes_the_comparison_group(self):
        self.section_template["remove_tag"] = True
        headers = compiler.eval_comparison_group_headers(
            self.columns, self.section_template, self.comparison_group
        )
        expected_headers = {"P ex1 vs ex2": "P", "A ex1 vs ex2": "A", "C ex1 vs ex2": "C"}  # fmt: skip
        assert headers == expected_headers

    def test_that_replace_comparison_tag_modifies_the_header(self):
        self.section_template["replace_comparison_tag"] = " / "
        headers = compiler.eval_comparison_group_headers(
            self.columns, self.section_template, self.comparison_group
        )
        expected_headers = {
            "P ex1 vs ex2": "P ex1 / ex2",
            "A ex1 vs ex2": "A ex1 / ex2",
            "C ex1 vs ex2": "C ex1 / ex2",
        }
        assert headers == expected_headers


@pytest.mark.parametrize(
    "replace_comparison_tag, expected_supheader",
    [(None, "A vs B"), (True, "A / B")],
)
def test_eval_comparison_group_supheader(replace_comparison_tag, expected_supheader):
    section_template = {"tag": " vs "}
    if replace_comparison_tag is not None:
        section_template["replace_comparison_tag"] = " / "
    supheader = compiler.eval_comparison_group_supheader(section_template, "A vs B")
    assert supheader == expected_supheader


def test_eval_comparison_group_conditional_format_names():
    columns = ["P ex1 vs ex2", "A ex1 vs ex2", "C ex1 vs ex2"]
    template_section = {"column_conditional": {"P": "cond_1", "A": "cond_2"}}
    col_conditionals = compiler.eval_comparison_group_conditional_format_names(columns, template_section)  # fmt: skip
    assert col_conditionals == {"P ex1 vs ex2": "cond_1", "A ex1 vs ex2": "cond_2"}


class TestEvalTagSampleHeaders:
    def test_with_remove_tag(self):
        template_section = {"tag": "Tag", "remove_tag": True}
        columns = ["Tag Sample 1", "Tag Sample 2"]
        expected = {"Tag Sample 1": "Sample 1", "Tag Sample 2": "Sample 2"}
        headers = compiler.eval_tag_sample_headers(columns, template_section)
        assert headers == expected

    def test_without_remove_tag(self):
        template_section = {"tag": "Tag", "remove_tag": False}
        columns = ["Tag Sample 1", "Tag Sample 2"]
        expected = {"Tag Sample 1": "Tag Sample 1", "Tag Sample 2": "Tag Sample 2"}
        headers = compiler.eval_tag_sample_headers(columns, template_section)
        assert headers == expected

    def test_without_remove_tag_and_log2_tag(self):
        template_section = {"tag": "Tag", "remove_tag": False, "log2": True}
        columns = ["Tag Sample 1", "Tag Sample 2"]
        expected = {
            "Tag Sample 1": "Tag Sample 1 [log2]",
            "Tag Sample 2": "Tag Sample 2 [log2]",
        }
        headers = compiler.eval_tag_sample_headers(columns, template_section, log2_tag="[log2]")  # fmt: skip
        assert headers == expected

    def test_with_remove_tag_and_log2_tag(self):
        template_section = {"tag": "Tag", "remove_tag": True, "log2": True}
        columns = ["Tag Sample 1", "Tag Sample 2"]
        expected = {
            "Tag Sample 1": "Sample 1",
            "Tag Sample 2": "Sample 2",
        }
        headers = compiler.eval_tag_sample_headers(columns, template_section, log2_tag="[log2]")  # fmt: skip
        assert headers == expected


class TestEvalTagSampleSupheader:
    def test_with_log2_tag(self):
        template_section = {"supheader": "Supheader", "log2": True, "tag": "Tag"}
        supheader = compiler.eval_tag_sample_supheader(template_section, log2_tag="[log2]")  # fmt: skip
        assert supheader == "Supheader [log2]"

    def test_without_log2_tag(self):
        template_section = {"supheader": "Supheader", "log2": False, "tag": "Tag"}
        supheader = compiler.eval_tag_sample_supheader(template_section, log2_tag="[log2]")  # fmt: skip
        assert supheader == "Supheader"

    def test_tag_is_used_if_supheader_not_specified(self):
        template_section = {"tag": "Tag as supheader"}
        supheader = compiler.eval_tag_sample_supheader(template_section, log2_tag="[log2]")  # fmt: skip
        assert supheader == "Tag as supheader"


class TestEvalColumnFormats:
    @pytest.fixture(autouse=True)
    def _init_inputs(self):
        self.columns = ["Column 1", "Column 2"]
        self.section_template = {"format": "str", "column_format": {"Column 1": "float"}}  # fmt: skip
        self.format_templates = {"str": {"align": "left"}, "float": {"align": "center"}}

    def test_with_only_section_format_specified(self):
        column_formats = compiler.eval_column_formats(
            self.columns, {"format": "str"}, self.format_templates
        )
        expected_formats = {
            "Column 1": self.format_templates["str"],
            "Column 2": self.format_templates["str"],
        }
        assert column_formats == expected_formats

    def test_with_section_and_column_format_specified(self):
        column_formats = compiler.eval_column_formats(
            self.columns, self.section_template, self.format_templates
        )
        expected_formats = {
            "Column 1": self.format_templates["float"],
            "Column 2": self.format_templates["str"],
        }
        assert column_formats == expected_formats

    def test_with_no_section_or_column_format_specified(self):
        column_formats = compiler.eval_column_formats(
            self.columns, {}, self.format_templates, default_format={"test": "test"}
        )
        expected_formats = {
            "Column 1": {"test": "test"},
            "Column 2": {"test": "test"},
        }
        assert column_formats == expected_formats

    def test_that_each_returned_format_is_a_unique_instance(self):
        column_formats = compiler.eval_column_formats(
            self.columns, {"format": "str"}, self.format_templates
        )
        assert column_formats["Column 1"] is not column_formats["Column 2"]
        assert column_formats["Column 1"] == column_formats["Column 2"]

    def test_border_true_adds_right_and_left_border_only_to_first_and_last_column(self):  # fmt: skip
        columns = ["Col 1", "Col 2", "Col 3"]
        column_formats = compiler.eval_column_formats(columns, {"border": True}, {})
        assert column_formats[columns[0]] == {"left": compiler.BORDER_TYPE}
        assert column_formats[columns[1]] == {}
        assert column_formats[columns[-1]] == {"right": compiler.BORDER_TYPE}

    def test_border_true_with_one_column_adds_left_and_right_border(self):
        self.section_template["border"] = True
        column_formats = compiler.eval_column_formats(["Col 1"], {"border": True}, {})
        assert column_formats["Col 1"] == {"left": compiler.BORDER_TYPE, "right": compiler.BORDER_TYPE}  # fmt: skip


class TestEvalColumnConditionalFormats:
    @pytest.fixture(autouse=True)
    def _init_inputs(self):
        self.columns = ["Column 1", "Column 2"]
        self.section_template = {"column_conditional": {"Column 1": "cond_1"}}
        self.conditional_format_templates = {"cond_1": {"type": "2_color_scale"}}

    def test_with_no_column_conditional_defined_in_the_section_template(self):
        column_formats = compiler.eval_column_conditional_formats(
            self.columns, {}, self.conditional_format_templates
        )
        expected_formats = {"Column 1": {}, "Column 2": {}}
        assert column_formats == expected_formats

    def test_with_column_conditional_defined_in_the_section_template(self):
        column_formats = compiler.eval_column_conditional_formats(
            self.columns, self.section_template, self.conditional_format_templates
        )
        expected_formats = {
            "Column 1": self.conditional_format_templates["cond_1"],
            "Column 2": {},
        }
        assert column_formats == expected_formats

    def test_with_no_conditional_defined_in_conditional_format_templates(self):
        column_formats = compiler.eval_column_conditional_formats(self.columns, self.section_template, {})  # fmt: skip
        expected_formats = {"Column 1": {}, "Column 2": {}}
        assert column_formats == expected_formats

    def test_that_each_returned_format_is_a_unique_instance(self):
        self.section_template["column_conditional"] = {"Column 1": "cond_1", "Column 2": "cond_1"}  # fmt: skip
        column_formats = compiler.eval_column_conditional_formats(
            self.columns, self.section_template, self.conditional_format_templates
        )
        assert column_formats["Column 1"] is not column_formats["Column 2"]
        assert column_formats["Column 1"] == column_formats["Column 2"]


class TestEvalColumnWidths:
    @pytest.fixture(autouse=True)
    def _init_inputs(self):
        self.columns = ["Column 1", "Column 2"]
        self.section_template = {"width": 70}

    def test_with_section_width_set(self):
        column_widths = compiler.eval_column_widths(self.columns, self.section_template)
        expected_widths = {"Column 1": 70, "Column 2": 70}
        assert column_widths == expected_widths

    def test_with_no_section_width_set(self):
        column_widths = compiler.eval_column_widths(self.columns, {}, default_width=0)
        expected_widths = {"Column 1": 0, "Column 2": 0}
        assert column_widths == expected_widths


class TestEvalHeaderFormats:
    @pytest.fixture(autouse=True)
    def _init_inputs(self):
        self.columns = ["Column 1", "Column 2"]
        self.section_template = {"header_format": {"bold": True}}
        self.format_templates = {"header": {"align": "center"}}

    def test_with_header_format_specified_in_section_and_format_template(self):
        header_formats = compiler.eval_header_formats(
            self.columns, self.section_template, self.format_templates
        )
        assert header_formats == {col: {"align": "center", "bold": True} for col in self.columns}  # fmt: skip

    def test_with_header_format_specified_in_section_but_not_in_format_template(self):
        header_formats = compiler.eval_header_formats(self.columns, self.section_template, {})  # fmt: skip
        assert header_formats == {col: {"bold": True} for col in self.columns}

    def test_with_header_format_not_specified_in_section_but_in_format_template(self):
        header_formats = compiler.eval_header_formats(self.columns, {}, self.format_templates)  # fmt: skip
        assert header_formats == {col: {"align": "center"} for col in self.columns}

    def test_with_header_format_not_specified_in_section_and_not_in_format_template(self):  # fmt: skip
        header_formats = compiler.eval_header_formats(self.columns, {}, {})
        assert header_formats == {col: {} for col in self.columns}

    def test_that_header_format_in_section_overwrites_the_header_format_template(self):
        header_formats = compiler.eval_header_formats(
            self.columns, {"header_format": {"a": 1}}, {"header": {"a": 2}}
        )
        assert header_formats == {col: {"a": 1} for col in self.columns}

    def test_that_each_returned_header_format_is_a_unique_instance(self):
        column_formats = compiler.eval_header_formats(
            self.columns, self.section_template, self.format_templates
        )
        assert column_formats["Column 1"] is not column_formats["Column 2"]
        assert column_formats["Column 1"] == column_formats["Column 2"]

    def test_that_original_formats_are_not_overwritten(self):
        original_template_format = self.format_templates["header"].copy()
        original_section_format = self.section_template["header_format"].copy()
        _ = compiler.eval_header_formats(
            self.columns, self.section_template, self.format_templates
        )
        assert original_template_format == self.format_templates["header"]
        assert original_section_format == self.section_template["header_format"]

    def test_border_true_adds_right_and_left_border_only_to_first_and_last_column(self):  # fmt: skip
        columns = ["Col 1", "Col 2", "Col 3"]
        header_formats = compiler.eval_header_formats(columns, {"border": True}, {})
        assert header_formats[columns[0]] == {"left": compiler.BORDER_TYPE}
        assert header_formats[columns[1]] == {}
        assert header_formats[columns[-1]] == {"right": compiler.BORDER_TYPE}

    def test_border_true_with_one_column_adds_left_and_right_border(self):
        self.section_template["border"] = True
        header_formats = compiler.eval_header_formats(
            ["Col 1"], {"format": "str", "border": True}, {"str": {}}
        )
        assert header_formats["Col 1"] == {"left": compiler.BORDER_TYPE, "right": compiler.BORDER_TYPE}  # fmt: skip


class TestEvalSupHeaderFormat:
    @pytest.fixture(autouse=True)
    def _init_inputs(self):
        self.section_template = {"supheader_format": {"bold": True}}
        self.format_templates = {"supheader": {"align": "center"}}

    def test_with_format_specified_in_section_and_format_template(self):
        sup_format = compiler.eval_supheader_format(self.section_template, self.format_templates)  # fmt: skip
        assert sup_format == {"align": "center", "bold": True}

    def test_with_format_specified_in_section_but_not_in_format_template(self):
        sup_format = compiler.eval_supheader_format(self.section_template, {})
        assert sup_format == {"bold": True}

    def test_with_format_not_specified_in_section_but_in_format_template(self):
        sup_format = compiler.eval_supheader_format({}, self.format_templates)
        assert sup_format == {"align": "center"}

    def test_with_format_not_specified_in_section_and_not_in_format_template(self):
        sup_format = compiler.eval_supheader_format({}, {})
        assert sup_format == {}

    def test_that_format_in_section_overwrites_the_header_format_template(self):
        sup_format = compiler.eval_supheader_format({"supheader_format": {"a": 1}}, {"supheader": {"a": 2}})  # fmt: skip
        assert sup_format == {"a": 1}

    def test_that_original_formats_are_not_overwritten(self):
        original_template_format = self.format_templates["supheader"].copy()
        original_section_format = self.section_template["supheader_format"].copy()
        _ = compiler.eval_supheader_format(self.section_template, self.format_templates)
        assert original_template_format == self.format_templates["supheader"]
        assert original_section_format == self.section_template["supheader_format"]

    def test_border_true(self):
        sup_format = compiler.eval_supheader_format({"border": True}, {})
        assert sup_format == {"left": compiler.BORDER_TYPE, "right": compiler.BORDER_TYPE}  # fmt: skip


class TestEvalSectionConditionalFormats:
    @pytest.fixture(autouse=True)
    def _init_inputs(self):
        self.section_template = {"conditional": "cond_1"}
        self.conditional_format_templates = {"cond_1": {"type": "2_color_scale"}}

    def test_with_no_column_conditional_defined_in_the_section_template(self):
        column_formats = compiler.eval_section_conditional_format({}, self.conditional_format_templates)  # fmt: skip
        expected_formats = {}
        assert column_formats == expected_formats

    def test_with_column_conditional_defined_in_the_section_template(self):
        column_formats = compiler.eval_section_conditional_format(self.section_template, self.conditional_format_templates)  # fmt: skip
        expected_formats = self.conditional_format_templates["cond_1"]
        assert column_formats == expected_formats

    def test_with_no_conditional_defined_in_conditional_format_templates(self):
        column_formats = compiler.eval_section_conditional_format(self.section_template, {})  # fmt: skip
        expected_formats = {}
        assert column_formats == expected_formats


def test_StandardSectionCompiler(
    report_template, example_table, standard_table_section
):
    section_compiler = compiler.StandardSectionCompiler(report_template)
    template_section = report_template.sections["Standard section 1"]
    compiled_section = section_compiler.compile(template_section, example_table)
    for attr in standard_table_section.__dataclass_fields__:
        if attr == "data":
            assert compiled_section.data.equals(standard_table_section.data)
        else:
            # Include attribute name in a dictionary to get nicer error messages
            compiled_attr = {attr: getattr(compiled_section, attr)}
            expected_section_attr = {attr: getattr(standard_table_section, attr)}
            assert compiled_attr == expected_section_attr


def test_TagSampleSectionCompiler(
    report_template, example_table, tag_sample_table_section
):
    section_compiler = compiler.TagSampleSectionCompiler(report_template)
    template_section = report_template.sections["Tag sample section 1"]
    compiled_section = section_compiler.compile(template_section, example_table)
    for attr in tag_sample_table_section.__dataclass_fields__:
        if attr == "data":
            assert compiled_section.data.equals(tag_sample_table_section.data)
        else:
            # Include attribute name in a dictionary to get nicer error messages
            compiled_attr = {attr: getattr(compiled_section, attr)}
            expected_section_attr = {attr: getattr(tag_sample_table_section, attr)}
            assert compiled_attr == expected_section_attr


class TestComparisonSectionCompiler:
    @pytest.fixture(autouse=True)
    def _init_inputs(self, report_template):
        self.table = pd.DataFrame(
            {
                "A column": [],
                "P ex1 vs ex2": [],
                "A ex1 vs ex2": [],
                "P ex1 vs EX3": [],
                "A ex1 vs EX3": [],
            }
        )
        self.section_template = {
            "comparison_group": True,
            "tag": " vs ",
            "columns": ["P", "A"],
            "column_conditional": {
                "P": "cond_1",
                "A": "cond_2",
            },
            "remove_tag": True,
            "replace_comparison_tag": " / ",
        }
        self.section_compiler = compiler.ComparisonSectionCompiler(report_template)

    def test_that_two_sections_with_correct_columns_are_generated(self):
        compiled_sections = self.section_compiler.compile(self.section_template, self.table)  # fmt: skip
        assert len(compiled_sections) == 2

        section_1, section_2 = compiled_sections
        assert section_1.data.columns.tolist() == ["P ex1 vs ex2", "A ex1 vs ex2"]
        assert section_2.data.columns.tolist() == ["P ex1 vs EX3", "A ex1 vs EX3"]

    def test_correct_application_of_conditional_formats(self):
        compiled_sections = self.section_compiler.compile(self.section_template, self.table)  # fmt: skip
        expected_column_conditionals = {
            "P ex1 vs ex2": {"type": "2_color_scale"},
            "A ex1 vs ex2": {"type": "3_color_scale"},
        }
        observed_column_conditionsls = compiled_sections[0].column_conditionals
        assert observed_column_conditionsls == expected_column_conditionals

    def test_compiled_sections_have_correct_headers(self):
        compiled_sections = self.section_compiler.compile(self.section_template, self.table)  # fmt: skip
        expected_headers = {"P ex1 vs ex2": "P", "A ex1 vs ex2": "A"}
        assert compiled_sections[0].headers == expected_headers

    def test_compiled_sections_have_correct_supheader(self):
        compiled_sections = self.section_compiler.compile(self.section_template, self.table)  # fmt: skip
        assert compiled_sections[0].supheader == "ex1 / ex2"
        assert compiled_sections[1].supheader == "ex1 / EX3"


class TestPrepareTableSections:
    def test_only_non_empty_sections_are_returned(self, report_template):
        table = pd.DataFrame({"Tag Sample 1": [1], "Tag Sample 2": [1]})
        compiled_sections = compiler.prepare_table_sections(report_template, table)
        assert all([not s.data.empty for s in compiled_sections])

    def test_duplicate_columns_are_removed(self, report_template, example_table):
        report_template.sections["Another section"] = {"columns": ["Column 1"]}
        compiled_sections = compiler.prepare_table_sections(report_template, example_table)  # fmt: skip
        observed_columns = []
        for section in compiled_sections:
            observed_columns.extend(section.data.columns)
        assert len(set(observed_columns)) == len(observed_columns)

    def test_duplicate_columns_are_kept_when_setting_is_false(self, report_template, example_table):  # fmt: skip
        report_template.sections["Another section"] = {"columns": ["Column 1"]}
        report_template.settings["remove_duplicate_columns"] = False
        compiled_sections = compiler.prepare_table_sections(report_template, example_table)  # fmt: skip
        assert compiled_sections[-1].data.columns.tolist() == ["Column 1"]

    def test_empty_sections_caused_by_removal_of_duplicate_columns_are_not_returned(
        self, report_template, example_table
    ):
        report_template.sections["Another section"] = {"columns": ["Column 1"]}
        compiled_sections = compiler.prepare_table_sections(report_template, example_table)  # fmt: skip
        assert all([not s.data.empty for s in compiled_sections])

    def test_addition_of_section_with_remaining_columns(self, report_template):  # fmt: skip
        table = pd.DataFrame({"Column 1": [1], "Another column": [1]})
        report_template.settings["append_remaining_columns"] = True
        compiled_sections = compiler.prepare_table_sections(report_template, table)
        assert len(compiled_sections) == 2


class TestCompileTableSection:
    def test_correctly_compiled_sections(
        self,
        report_template,
        example_table,
        standard_table_section,
        tag_sample_table_section,
    ):
        compiled_sections = compiler.compile_table_sections(report_template, example_table)  # fmt: skip
        expected_sections = [standard_table_section, tag_sample_table_section]
        for compiled_section, expected_section in zip(expected_sections, compiled_sections):  # fmt: skip
            for attr in expected_section.__dataclass_fields__:
                if attr == "data":
                    assert compiled_section.data.equals(expected_section.data)
                else:
                    # Include attribute name in a dictionary to get nicer error messages
                    compiled_attr = {attr: getattr(compiled_section, attr)}
                    expected_section_attr = {attr: getattr(expected_section, attr)}
                    assert compiled_attr == expected_section_attr

    @pytest.mark.parametrize(
        "section",
        [
            {"format": "str"},
            {"columns": ["Column 1", "Column 2"], "comparison_group": True},
            {"columns": ["Column 1", "Column 2"], "tag": "str"},
        ],
    )
    def test_invalid_sections_are_not_compiled(
        self, report_template, example_table, section
    ):
        report_template.sections = {"invalid": section}
        compiled_sections = compiler.compile_table_sections(report_template, example_table)  # fmt: skip
        assert len(compiled_sections) == 0


class TestCompileRemaininColumnTableSection:
    def test_correct_columns_selected_for_section(
        self, report_template, example_table, standard_table_section
    ):
        compiled_section = compiler.compile_remaining_column_table_section(
            report_template, [standard_table_section], example_table
        )
        expected_columns = [
            c for c in example_table if c not in standard_table_section.data
        ]
        assert not standard_table_section.data.columns.empty
        assert compiled_section.data.columns.tolist() == expected_columns

    def test_empty_section_returned_when_no_remaining_columns(
        self, report_template, example_table, standard_table_section
    ):
        example_table = example_table[standard_table_section.data.columns]
        compiled_section = compiler.compile_remaining_column_table_section(
            report_template, [standard_table_section], example_table
        )
        assert compiled_section.data.empty


class TestPruneTableSections:
    def test_duplicate_columns_removed_from_latter_sections(self):
        sections = [
            compiler.TableSection(data=pd.DataFrame(columns=["C1", "C2"])),
            compiler.TableSection(data=pd.DataFrame(columns=["C1", "C3"])),
            compiler.TableSection(data=pd.DataFrame(columns=["C1", "C4"])),
        ]
        compiler.prune_table_sections(sections)
        assert sections[0].data.columns.tolist() == ["C1", "C2"]
        assert sections[1].data.columns.tolist() == ["C3"]
        assert sections[2].data.columns.tolist() == ["C4"]

    def test_removal_of_all_columns_from_a_section_returns_an_empty_section(self):
        sections = [
            compiler.TableSection(data=pd.DataFrame(columns=["C1", "C2"])),
            compiler.TableSection(data=pd.DataFrame(columns=["C1", "C2"])),
        ]
        compiler.prune_table_sections(sections)
        assert sections[0].data.columns.tolist() == ["C1", "C2"]
        assert sections[1].data.columns.tolist() == []

    def test_unique_columns_not_removed(self):
        sections = [
            compiler.TableSection(data=pd.DataFrame(columns=["C1", "C2"])),
            compiler.TableSection(data=pd.DataFrame(columns=["C3", "C4"])),
        ]
        compiler.prune_table_sections(sections)
        for section in sections:
            assert len(section.data.columns) == 2
            assert len(sections[1].column_formats) == 2
            assert len(sections[1].column_conditionals) == 2
            assert len(sections[1].column_widths) == 2
            assert len(sections[1].headers) == 2
            assert len(sections[1].header_formats) == 2

    def test_column_removed_from_all_section_parameters(self):
        sections = [
            compiler.TableSection(data=pd.DataFrame(columns=["C1", "C2"])),
            compiler.TableSection(data=pd.DataFrame(columns=["C1", "C3"])),
        ]
        compiler.prune_table_sections(sections)
        assert sections[1].data.columns.tolist() == ["C3"]
        assert list(sections[1].column_formats) == ["C3"]
        assert list(sections[1].column_conditionals) == ["C3"]
        assert list(sections[1].column_widths) == ["C3"]
        assert list(sections[1].headers) == ["C3"]
        assert list(sections[1].header_formats) == ["C3"]


def test_remove_empty_table_sections():
    sections = [
        compiler.TableSection(data=pd.DataFrame({"C1": [1]})),
        compiler.TableSection(data=pd.DataFrame({})),
        compiler.TableSection(data=pd.DataFrame({"C2": [1]})),
    ]
    filtered_sections = compiler.remove_empty_table_sections(sections)
    assert all([not s.data.empty for s in filtered_sections])


class TestIdentifyTemplateSectionCategory:
    def test_identify_standard_section(self):
        section_template = {"columns": []}
        section_category = compiler.identify_template_section_category(section_template)
        assert section_category == compiler.SectionCategory.STANDARD

    def test_identify_tag_sample_section(self):
        section_template = {"tag": ""}
        section_category = compiler.identify_template_section_category(section_template)
        assert section_category == compiler.SectionCategory.TAG_SAMPLE

    @pytest.mark.parametrize(
        "section_template",
        [{"comparison_group": True, "columns": [], "tag": ""}],
    )
    def test_identify_comparison_section(self, section_template):
        section_category = compiler.identify_template_section_category(section_template)
        assert section_category == compiler.SectionCategory.COMPARISON

    @pytest.mark.parametrize(
        "section_template",
        [
            {"format": "str"},
            {"columns": [], "tag": ""},
            {"columns": [], "comparison_group": True},
            {"tag": "", "comparison_group": True},
            {"columns": [], "tag": "", "comparison_group": False},
        ],
    )
    def test_identify_unknown_section(self, section_template):
        section_category = compiler.identify_template_section_category(section_template)
        assert section_category == compiler.SectionCategory.UNKNOWN



"""This module provides a class for handling configuration for formatting excel reports.

The module contains a single class, ReportTemplate, which stores the configuration of a
report. It provides methods for loading and saving the configuration to a YAML file, as
well as checking if the configuration is valid.

Valid ReportTemplate.settings entries are:
    supheader_height: float (default: 20)
    header_height: float (default: 20)
    column_width: float (default: 64)
    log2_tag: str
    sample_extraction_tag: str
    append_remaining_columns: bool (default: False)
    write_supheader: bool (default: True)
    evaluate_log2_transformation: bool (default: False)
    remove_duplicate_columns: bool (default: True)
    add_autofilter: bool (default: True)
    freeze_cols: int (default: 1)

"""

from __future__ import annotations
from typing import Optional

import yaml


class ReportTemplate:
    """Class to store the template of a report."""

    def __init__(
        self,
        sections: Optional[dict] = None,
        formats: Optional[dict] = None,
        settings: Optional[dict] = None,
        conditional_formats: Optional[dict] = None,
    ):
        self.sections = {} if sections is None else sections
        self.formats = {} if formats is None else formats
        self.settings = {} if settings is None else settings
        self.conditional_formats = (
            {} if conditional_formats is None else conditional_formats
        )

    @classmethod
    def load(cls, filepath) -> ReportTemplate:
        """Load a report template from a YAML file."""
        with open(filepath, "r", encoding="utf-8") as file:
            template_data = yaml.safe_load(file)
        template = cls()
        template.formats = template_data.get("formats", {})
        template.conditional_formats = template_data.get("conditional_formats", {})
        template.sections = template_data.get("groups", {})
        template.settings = template_data.get("args", {})
        return template

    def save(self, filepath) -> None:
        """Save a report template to a YAML file."""
        template_data = {
            "groups": self.sections,
            "formats": self.formats,
            "conditional_formats": self.conditional_formats,
            "args": self.settings,
        }
        with open(filepath, "w", encoding="utf-8") as file:
            yaml.dump(
                template_data,
                file,
                version=(1, 2),
                sort_keys=False,
                Dumper=IndentDumper,
            )


class IndentDumper(yaml.SafeDumper):
    """Custom YAML dumper to preserve indentation."""

    def increase_indent(self, flow=False, indentless=False):
        return super(IndentDumper, self).increase_indent(flow, False)



import os
import warnings

import click
import pandas as pd

import xlsxreport


@click.command()
@click.argument("infile")
@click.option(
    "--config",
    default="cassiopeia.yaml",
    help="Name of the config file, default is 'cassiopeia.yaml'.",
)
@click.option(
    "--outfile",
    default="cassiopeia_report.xlsx",
    help="Name of the report file, default is 'cassiopeia_report.xlsx'.",
)
@click.option(
    "--outpath",
    default=None,
    help=(
        "Output path of the report file, if specified overrides the outfile parameter."
    ),
)
@click.option(
    "--sep", default="\t", help="Delimiter to use for 'infile', default is \\t."
)
def cli(infile: str, config: str, outfile: str, outpath: str, sep: str) -> None:
    """Description of the XlsxReport function."""
    if outpath is not None:
        report_path = outpath
    else:
        report_path = os.path.join(os.path.dirname(infile), outfile)

    if os.path.isfile(config):
        config_path = config
    else:
        config_path = xlsxreport.get_config_file(config)

    if not os.path.isfile(config_path):
        click.echo("Config file not found: %s" % config_path)

    click.echo(f"\n")
    click.echo(f"Preparing xlsx report:")
    click.echo(f"----------------------")
    click.echo(f"\tFile:   {infile}")
    click.echo(f"\tConfig: {config_path}\n")

    with warnings.catch_warnings():
        warnings.simplefilter(action="ignore", category=pd.errors.DtypeWarning)
        table = pd.read_csv(infile, sep=sep)

    # Replace column names - adds a "total" keyword to intensity columns
    replace_column_names = {
        "Intensity": "Intensity total",
        "norm imp intensity": "norm imp intensity total",
        "iBAQ": "iBAQ total",
        "norm intensity": "norm intensity total",
        "LFQ intensity": "LFQ intensity total",
    }
    table.rename(columns=replace_column_names, inplace=True)

    # Replace comparison column names
    replace_column_tags = [
        ("P.Value_", "P.Value"),
        ("adj.P.Val_", "adj.P.Val"),
        ("logFC_", "logFC"),
        ("AveExpr_", "AveExpr"),
    ]
    for old_tag, new_tag in replace_column_tags:
        table.columns = [c.replace(old_tag, new_tag) for c in table.columns]

    # Sort rows
    try:
        sort_rows_by = "MS/MS count"
        sort_ascending = [False]
        table.sort_values(
            sort_rows_by,
            ascending=sort_ascending,
            na_position="first",
            inplace=True,
        )
    except KeyError:
        pass

    # Generate excel report
    with xlsxreport.Reportbook(report_path) as reportbook:
        protein_sheet = reportbook.add_datasheet("Proteins")
        protein_sheet.apply_configuration(config_path)
        protein_sheet.add_data(table)
        protein_sheet.write_data()
    click.echo(f"\tReport written to:")
    click.echo(f"\t{report_path}")


if __name__ == "__main__":
    cli()



import os

import pandas as pd
import pytest
import openpyxl
import xlsxwriter

import xlsxreport
from xlsxreport.excel_writer import TableSectionWriter
from xlsxreport.template import ReportTemplate
from xlsxreport.compiler import prepare_table_sections


@pytest.fixture()
def temp_excel_path(request, tmp_path):
    output_path = os.path.join(tmp_path, "output.xlsx")

    def teardown():
        if os.path.isfile(output_path):
            os.remove(output_path)

    request.addfinalizer(teardown)
    return output_path


class TestCorrectCreationOfFormattedExcelFile:
    def test_old_xlsx_report_implementation(self, temp_excel_path):
        config_path = "./tests/testdata/mq_protein_config.yaml"
        mq_path = "./tests/testdata/mq_proteinGroups.txt"
        table = pd.read_csv(mq_path, sep="\t")
        with xlsxreport.Reportbook(temp_excel_path) as reportbook:
            protein_sheet = reportbook.add_datasheet("Proteins")
            protein_sheet.apply_configuration(config_path)
            protein_sheet.add_data(table)
            protein_sheet.write_data()

        reference_excel_path = "./tests/testdata/mq_proteinGroups.xlsx"
        with open(reference_excel_path, "rb") as f1, open(temp_excel_path, "rb") as f2:
            wb1 = openpyxl.load_workbook(f1)
            wb2 = openpyxl.load_workbook(f2)

            # Compare the number of sheets
            assert len(wb1.sheetnames) == len(wb2.sheetnames)

            # Compare the sheet names
            for sheet1, sheet2 in zip(wb1, wb2):
                assert sheet1.title == sheet2.title

            # Compare equal number of rows and columns
            assert len(list(sheet1.iter_rows())) == len(list(sheet2.iter_rows()))
            assert len(list(sheet1.iter_cols())) == len(list(sheet2.iter_cols()))

            # Compare the cell values and formatting
            for sheet1, sheet2 in zip(wb1, wb2):
                for row1, row2 in zip(sheet1.iter_rows(), sheet2.iter_rows()):
                    for cell1, cell2 in zip(row1, row2):
                        assert cell1.value == cell2.value
                        assert cell1._style == cell2._style
                        assert str(cell1.number_format) == str(cell2.number_format)
                        assert str(cell1.font) == str(cell2.font)
                        assert str(cell1.border) == str(cell2.border)
                        assert str(cell1.fill) == str(cell2.fill)
                        assert str(cell1.alignment) == str(cell2.alignment)

            # Compare the conditional formatting rules
            for sheet1, sheet2 in zip(wb1, wb2):
                for rule1, rule2 in zip(
                    sheet1.conditional_formatting._cf_rules,
                    sheet2.conditional_formatting._cf_rules,
                ):
                    assert rule1.cells == rule2.cells
                    for cfrule1, cfrul2 in zip(rule1.rules, rule2.rules):
                        assert str(cfrule1) == str(cfrul2)

    def test_new_xlsx_report_implementation(self, temp_excel_path):
        config_path = "./tests/testdata/mq_protein_config.yaml"
        mq_path = "./tests/testdata/mq_proteinGroups.txt"
        table = pd.read_csv(mq_path, sep="\t")

        report_template = ReportTemplate.load(config_path)
        table_sections = prepare_table_sections(report_template, table)
        with xlsxwriter.Workbook(temp_excel_path) as workbook:
            worksheet = workbook.add_worksheet("Proteins")
            section_writer = TableSectionWriter(workbook)
            section_writer.write_sections(
                worksheet, table_sections, settings=report_template.settings
            )

        reference_excel_path = "./tests/testdata/mq_proteinGroups.xlsx"
        with open(reference_excel_path, "rb") as f1, open(temp_excel_path, "rb") as f2:
            wb1 = openpyxl.load_workbook(f1)
            wb2 = openpyxl.load_workbook(f2)

            # Compare the number of sheets
            assert len(wb1.sheetnames) == len(wb2.sheetnames)

            # Compare the sheet names
            for sheet1, sheet2 in zip(wb1, wb2):
                assert sheet1.title == sheet2.title

            # Compare equal number of rows and columns
            assert len(list(sheet1.iter_rows())) == len(list(sheet2.iter_rows()))
            assert len(list(sheet1.iter_cols())) == len(list(sheet2.iter_cols()))

            # Compare the cell values and formatting
            for sheet1, sheet2 in zip(wb1, wb2):
                for row1, row2 in zip(sheet1.iter_rows(), sheet2.iter_rows()):
                    for cell1, cell2 in zip(row1, row2):
                        assert cell1.value == cell2.value
                        assert cell1._style == cell2._style
                        assert str(cell1.number_format) == str(cell2.number_format)
                        assert str(cell1.font) == str(cell2.font)
                        assert str(cell1.border) == str(cell2.border)
                        assert str(cell1.fill) == str(cell2.fill)
                        assert str(cell1.alignment) == str(cell2.alignment)

            # Compare the conditional formatting rules
            for sheet1, sheet2 in zip(wb1, wb2):
                for rule1, rule2 in zip(
                    sheet1.conditional_formatting._cf_rules,
                    sheet2.conditional_formatting._cf_rules,
                ):
                    assert rule1.cells == rule2.cells
                    for cfrule1, cfrul2 in zip(rule1.rules, rule2.rules):
                        assert str(cfrule1) == str(cfrul2)



"""Contains functions for compiling table sections from a report template and a table."""
from __future__ import annotations
from dataclasses import dataclass, field
from enum import Enum
from typing import Iterable, Optional, Protocol, Sequence

import numpy as np
import pandas as pd

from xlsxreport.template import ReportTemplate


BORDER_TYPE: int = 2  # 2 = thick line, see xlsxwriter.format.Format().set_border()
DEFAULT_COL_WIDTH: float = 64
DEFAULT_FORMAT: dict = {"num_format": "@"}
REMAINING_COL_FORMAT = {"align": "left", "num_format": "0"}
NAN_REPLACEMENT_SYMBOL = ""
WHITESPACE_CHARS = " ."


class SectionCategory(Enum):
    """Enum for section categories."""

    UNKNOWN = -1
    STANDARD = 1
    TAG_SAMPLE = 2
    COMPARISON = 3


@dataclass
class TableSection:
    """Contains information for writing and formatting a section of a table.

    Note that the `data` DataFrame must not contain any NaN values.
    """

    data: pd.DataFrame
    column_formats: dict = field(default_factory=dict)
    column_conditionals: dict = field(default_factory=dict)
    column_widths: dict = field(default_factory=dict)
    headers: dict = field(default_factory=dict)
    header_formats: dict = field(default_factory=dict)
    supheader: str = ""
    supheader_format: dict = field(default_factory=dict)
    section_conditional: dict = field(default_factory=dict)
    hide_section: bool = False

    def __post_init__(self):
        nan_columns = self.data.columns[self.data.isnull().any()].tolist()
        if nan_columns:
            raise ValueError(f"`data` contains NaN values in columns: {nan_columns}")

        for col in self.data.columns:
            if col not in self.column_formats:
                self.column_formats[col] = {}
            if col not in self.column_conditionals:
                self.column_conditionals[col] = {}
            if col not in self.column_widths:
                self.column_widths[col] = DEFAULT_COL_WIDTH
            if col not in self.headers:
                self.headers[col] = col
            if col not in self.header_formats:
                self.header_formats[col] = {}


class SectionCompiler(Protocol):
    """Protocol for section compilers."""

    def __init__(self, report_template: ReportTemplate):
        ...

    def compile(
        self, section_template: dict, table: pd.DataFrame
    ) -> TableSection | list[TableSection]:
        """Compile a table section from a section template and a table."""
        ...


class StandardSectionCompiler:
    """Compiler for standard table sections."""

    def __init__(self, report_template: ReportTemplate):
        self.formats = report_template.formats
        self.conditional_formats = report_template.conditional_formats
        self.settings = report_template.settings

    def compile(self, section_template: dict, table: pd.DataFrame) -> TableSection:
        """Compile a table section from a standard section template and a table."""
        selected_cols = eval_standard_section_columns(table.columns, section_template)
        data = eval_data(table, selected_cols)
        col_formats = eval_column_formats(
            selected_cols, section_template, self.formats, DEFAULT_FORMAT
        )
        col_conditionals = eval_column_conditional_formats(
            selected_cols, section_template, self.conditional_formats
        )
        default_width = self.settings.get("column_width", DEFAULT_COL_WIDTH)
        col_widths = eval_column_widths(selected_cols, section_template, default_width)
        headers = {c: c for c in selected_cols}
        header_formats = eval_header_formats(
            selected_cols, section_template, self.formats
        )
        supheader = section_template.get("supheader", "")
        supheader_format = eval_supheader_format(section_template, self.formats)
        section_conditional = eval_section_conditional_format(
            section_template, self.conditional_formats
        )
        hide_section = section_template.get("hide_section", False)

        return TableSection(
            data=data,
            column_formats=col_formats,
            column_conditionals=col_conditionals,
            column_widths=col_widths,
            headers=headers,
            header_formats=header_formats,
            supheader=supheader,
            supheader_format=supheader_format,
            section_conditional=section_conditional,
            hide_section=hide_section,
        )


class TagSampleSectionCompiler:
    """Compiler for tag sample table sections."""

    def __init__(self, report_template: ReportTemplate):
        self.formats = report_template.formats
        self.conditional_formats = report_template.conditional_formats
        self.settings = report_template.settings

    def compile(self, section_template: dict, table: pd.DataFrame) -> TableSection:
        """Compile a table section from a standard section template and a table."""
        selected_cols = eval_tag_sample_section_columns(
            table.columns, section_template, self.settings["sample_extraction_tag"]
        )
        data = eval_data_with_log2_transformation(
            table,
            selected_cols,
            section_template,
            self.settings.get("evaluate_log2_transformation", False),
        )
        col_formats = eval_column_formats(
            selected_cols, section_template, self.formats, DEFAULT_FORMAT
        )
        col_conditionals = eval_column_conditional_formats(
            selected_cols, section_template, self.conditional_formats
        )
        default_width = self.settings.get("column_width", DEFAULT_COL_WIDTH)
        col_widths = eval_column_widths(selected_cols, section_template, default_width)
        headers = eval_tag_sample_headers(
            selected_cols, section_template, self.settings.get("log2_tag", "")
        )
        header_formats = eval_header_formats(
            selected_cols, section_template, self.formats
        )
        supheader = eval_tag_sample_supheader(
            section_template, self.settings.get("log2_tag", "")
        )
        supheader_format = eval_supheader_format(section_template, self.formats)
        section_conditional = eval_section_conditional_format(
            section_template, self.conditional_formats
        )
        hide_section = section_template.get("hide_section", False)

        return TableSection(
            data=data,
            column_formats=col_formats,
            column_conditionals=col_conditionals,
            column_widths=col_widths,
            headers=headers,
            header_formats=header_formats,
            supheader=supheader,
            supheader_format=supheader_format,
            section_conditional=section_conditional,
            hide_section=hide_section,
        )


class ComparisonSectionCompiler:
    """Compiler for comparison table sections."""

    def __init__(self, report_template: ReportTemplate):
        self.standard_compiler = StandardSectionCompiler(report_template)

    def compile(
        self, section_template: dict, table: pd.DataFrame
    ) -> list[TableSection]:
        """Compile table sections from a comparison section template and a table."""

        comparison_groups = eval_comparison_groups(table.columns, section_template)
        table_sections = []
        for comparison_group in comparison_groups:
            selected_cols = eval_comparison_group_columns(
                table.columns, section_template, comparison_group
            )
            col_conditionals = eval_comparison_group_conditional_format_names(
                selected_cols, section_template
            )
            supheader = eval_comparison_group_supheader(
                section_template, comparison_group
            )

            std_section_template = section_template.copy()
            std_section_template["columns"] = selected_cols
            std_section_template["column_conditional"] = col_conditionals
            std_section_template["supheader"] = supheader

            table_section = self.standard_compiler.compile(std_section_template, table)
            table_section.headers = eval_comparison_group_headers(
                selected_cols, section_template, comparison_group
            )
            table_sections.append(table_section)

        return table_sections


def get_section_compiler(section_template: dict) -> type[SectionCompiler]:
    """Get the section compiler function for a section template."""
    section_category = identify_template_section_category(section_template)
    if section_category == SectionCategory.UNKNOWN:
        raise ValueError("Unknown section category.")
    elif section_category == SectionCategory.STANDARD:
        return StandardSectionCompiler
    elif section_category == SectionCategory.TAG_SAMPLE:
        return TagSampleSectionCompiler
    elif section_category == SectionCategory.COMPARISON:
        return ComparisonSectionCompiler
    else:
        raise NotImplementedError(
            f"Section compiler not implemented for category {section_category}."
        )


def prepare_table_sections(
    report_template: ReportTemplate,
    table: pd.DataFrame,
) -> list[TableSection]:
    """Compile non-empty table sections from a report template and a table.

    Args:
        report_template: The report template describing how table sections should be
            generated.
        table: The table to compile the sections from.

    Returns:
        A list of non-empty, compiled table sections.
    """
    compiled_table_sections = compile_table_sections(report_template, table)
    if report_template.settings.get("append_remaining_columns", False):
        remaining_section = compile_remaining_column_table_section(
            report_template, compiled_table_sections, table
        )
        compiled_table_sections.append(remaining_section)
    if report_template.settings.get("remove_duplicate_columns", True):
        prune_table_sections(compiled_table_sections)
    return remove_empty_table_sections(compiled_table_sections)


def compile_table_sections(
    report_template: ReportTemplate, table: pd.DataFrame
) -> list[TableSection]:
    """Compile table sections from a report template and a table.

    Args:
        report_template: The report template describing how table sections should be
            generated.
        table: The table to compile the sections from.

    Returns:
        A list of compiled table sections.
    """
    table_sections = []
    for section_template in report_template.sections.values():
        section_category = identify_template_section_category(section_template)
        if section_category == SectionCategory.UNKNOWN:
            continue

        SectionCompilerClass = get_section_compiler(section_template)
        section_compiler = SectionCompilerClass(report_template)
        compiled_section = section_compiler.compile(section_template, table)
        if isinstance(compiled_section, TableSection):
            table_sections.append(compiled_section)
        else:
            table_sections.extend(compiled_section)

    return table_sections


def compile_remaining_column_table_section(
    report_template: ReportTemplate,
    table_sections: Iterable[TableSection],
    table: pd.DataFrame,
) -> TableSection:
    """Compile a table section containing all columns not present in other sections.

    Args:
        report_template: The report template describing how table sections should be
            generated.
        table_sections: The table sections that have already been compiled.
        table: The table to compile the remaining column section from.

    Returns:
        A compiled table section containing all columns not present in other sections.
    """
    observed_columns = set()
    for section in table_sections:
        observed_columns.update(section.data.columns)
    selected_cols = [column for column in table if column not in observed_columns]

    section_compiler = StandardSectionCompiler(report_template)
    section_compiler.formats[-1] = REMAINING_COL_FORMAT
    section_template = {
        "columns": selected_cols,
        "format": -1,
        "width": DEFAULT_COL_WIDTH,
        "hide_section": True,
    }
    section = section_compiler.compile(section_template, table)
    del section_compiler.formats[-1]
    return section


def prune_table_sections(table_sections: Iterable[TableSection]) -> None:
    """Remove duplicate columns from table sections, keeping only the first occurance."""
    observed_columns = set()
    for section in table_sections:
        to_remove = [col for col in section.data.columns if col in observed_columns]
        section.data = section.data.drop(columns=to_remove)
        for col in to_remove:
            del section.column_formats[col]
            del section.column_conditionals[col]
            del section.column_widths[col]
            del section.headers[col]
            del section.header_formats[col]
        observed_columns.update(section.data.columns)


def remove_empty_table_sections(
    table_sections: Iterable[TableSection],
) -> list[TableSection]:
    """Returns a list of non-empty table sections."""
    return [section for section in table_sections if not section.data.empty]


def eval_data(table: pd.DataFrame, columns: Iterable[str]) -> pd.DataFrame:
    """Returns a copy of the table with only the selected columns and no NaN values.

    Args:
        table: The table to select columns from.
        columns: The columns to select from the table.

    Returns:
        A copy of the table with only the selected columns NaN values replaced.
    """
    data = table[columns].astype("object").fillna(NAN_REPLACEMENT_SYMBOL)
    return data


def eval_data_with_log2_transformation(
    table: pd.DataFrame,
    columns: Iterable[str],
    section_template: dict,
    evaluate_log_state: bool,
) -> pd.DataFrame:
    """Selects columns from the table and applies a log2 transformation if specified.

    Args:
        table: The table to select columns from.
        columns: The columns to select from the table.
        section_template: A dictionary containing the key "log2" (bool), which
            determines whether to apply a log2 transformation.
        evaluate_log_state: If True, the log2 transformation is only applied if the
            intensities are not already in log space.

    Returns:
        A copy of the table with only the selected columns and NaN vaues replaced,
        optionally values are log2 transformed.
    """
    data = table[columns].copy()
    apply_log_transformation = section_template.get("log2", False)
    if evaluate_log_state and _intensities_in_logspace(data):
        apply_log_transformation = False

    if apply_log_transformation:
        if not data.select_dtypes(exclude=["number"]).columns.empty:
            raise ValueError("Cannot log2 transform non-numeric columns.")
        data = data.mask(data <= 0, np.nan)
        data = np.log2(data)
    data = data.astype("object").fillna(NAN_REPLACEMENT_SYMBOL)
    return data


def eval_standard_section_columns(
    columns: Iterable[str], section_template: dict
) -> list[str]:
    """Select columns from the template that are present in the table.

    Args:
        columns: A list of column names to select from.
        section_template: A dictionary containing the columns to be selected as the
            values of the "columns" key.

    Returns:
        A list of column names that are present in both the template and the table.
    """
    selected_columns = [col for col in section_template["columns"] if col in columns]
    return selected_columns


def eval_tag_sample_section_columns(
    columns: Iterable[str], section_template: dict, extraction_tag: str
) -> list[str]:
    """Extract tag sample columns.

    Args:
        columns: A list of column names to select from.
        section_template: A dictionary containing the columns to be selected as the
            values of the "columns" key.
        extraction_tag: The tag used to extract sample names from the columns.

    Returns:
        A list of sample columns that contain the `section_template["tag"]`.
    """
    samples = []
    for col in columns:
        if extraction_tag not in col or col == extraction_tag:
            continue
        samples.append(col.replace(extraction_tag, "").strip(WHITESPACE_CHARS))

    selected_columns = []
    for col in columns:
        sample_query = col.replace(section_template["tag"], "").strip(WHITESPACE_CHARS)
        if section_template["tag"] in col and sample_query in samples:
            selected_columns.append(col)
    return selected_columns


def eval_comparison_groups(columns: Iterable[str], section_template: dict) -> list[str]:
    """Extract comparison groups from the columns of a table.

    Args:
        columns: A list of column names used for extracting comparison group names.
        section_template: A dictionary containing the keys "tag" (a string) and
            "columns" (a list of strings).

    Returns:
        A list of unique comparison group names. Comparison group names are extracted
        from the `columns` that contain the substring specified in the
        `section_template` by "tag" and one of the substrings specified by "columns".
        Comparison groups are extracted by removing the `section_template["columns"]`
        substrings from the column name and stripping whitespace characters from the
        result. Each comparison group name is extracted only once.
    """
    comparison_tag = section_template["tag"]
    comparison_columns = [col for col in columns if comparison_tag in col]

    comparison_groups = []
    for column_tag in section_template["columns"]:
        matched_columns = [col for col in comparison_columns if column_tag in col]
        for column in matched_columns:
            putative_group = column.replace(column_tag, "").strip(WHITESPACE_CHARS)
            if putative_group and putative_group not in comparison_groups:
                comparison_groups.append(putative_group)
    return comparison_groups


def eval_comparison_group_columns(
    columns: Iterable[str], section_template: dict, comparison_group: str
) -> list[str]:
    """Extract columns from a table that belong to a comparison group.

    Args:
        columns: A list of column names to select from.
        section_template: A dictionary containing the key "columns" with its value
            being a list of strings.
        comparison_group: The name of the comparison group to extract columns for.

    Returns:
        A list of column names that consist of the `comparison_group` and one of the
        substrings specified by `section_template["columns"]`.
    """
    selected_cols = []
    comparison_group_cols = [col for col in columns if comparison_group in col]
    for column_tag in section_template["columns"]:
        for column in comparison_group_cols:
            leftover = column.replace(comparison_group, "").replace(column_tag, "")
            if leftover.strip(WHITESPACE_CHARS) == "" and column not in selected_cols:
                selected_cols.append(column)
                break
    return selected_cols


def eval_comparison_group_headers(
    columns: Iterable[str], section_template: dict, comparison_group: str
) -> dict[str, str]:
    """Returns header names for each column.

    Args:
        columns: A list of column names to select from.
        section_template: A dictionary with keys "remove_tag" (bool), "tag" (str), and
            "replace_comparison_tag" (str). If "remove_tag" is True, the
            `comparison_group` will be removed from the headers. If the
            "replace_comparison_tag" is specified, the "tag" value will be replaced with
            the value of "replace_comparison_tag" in the headers.
        comparison_group: The name of the comparison group.

    Returns:
        A dictionary containing the header names for each column.
    """
    remove_comparison_group = section_template.get("remove_tag", False)
    old_comparison_tag = section_template["tag"]
    new_comparison_tag = section_template.get("replace_comparison_tag", None)
    replace_comparison_tag = True if new_comparison_tag is not None else False

    headers = {}
    for col in columns:
        header = col
        if remove_comparison_group:
            header = header.replace(comparison_group, "").strip(WHITESPACE_CHARS)
        if replace_comparison_tag:
            header = header.replace(old_comparison_tag, new_comparison_tag)
        headers[col] = header
    return headers


def eval_comparison_group_supheader(
    section_template: dict, comparison_group: str
) -> str:
    """Returns the supheader name for a comparison group.

    Args:
        section_template: A dictionary with keys "tag" (str) and
            "replace_comparison_tag" (str). If the "replace_comparison_tag" is
            specified, the "tag" value will be replaced with the value of
            "replace_comparison_tag" in the supheader.
        comparison_group: The name used for the supheader.

    Returns:
        The supheader name for the comparison group section.
    """
    old_comparison_tag = section_template["tag"]
    new_comparison_tag = section_template.get("replace_comparison_tag", None)
    replace_comparison_tag = True if new_comparison_tag is not None else False

    supheader = comparison_group
    if replace_comparison_tag:
        supheader = supheader.replace(old_comparison_tag, new_comparison_tag)
    return supheader


def eval_comparison_group_conditional_format_names(
    columns: Iterable[str], section_template: dict
) -> dict[str, str]:
    """Returns conditional format names for each column.

    Conditional formats are matched to each column by checking if any of the
    `section_template["conditional format"]` keys are a substring of the column name,
    and using the corresponding value as the conditional format name for the column.

    Args:
        columns: A list of column names.
        section_template: A dictionary with a "column_conditional" key containing a
            dictionary with column names as keys and conditional format names as values.

    Returns:
        A dictionary containing conditional format names for each column.
    """
    conditional_formats = section_template.get("column_conditional", {})
    col_conditionals = {}
    for tag, format_name in conditional_formats.items():
        for column in columns:
            if tag in column:
                col_conditionals[column] = format_name
    return col_conditionals


def eval_tag_sample_headers(
    columns: Iterable[str],
    section_template: dict,
    log2_tag: str = "",
) -> dict[str, str]:
    """Returns header names for each column.

    Args:
        columns: A list of column names to select from.
        section_template: A dictionary with "tag" containing the substring that will be
            removed from the headers if "remove_tag" is True. The "log2" key determines
            whether to add the `log2_tag` to the headers, however, if "remove_tag" is
            True the `log2_tag` will never be added. The "remove_tag" and "log"
            keys are optional and by default False.
        log2_tag: The substring that will be added to the column names if `log2` is
            True.

    Returns:
        A dictionary containing the header names for each column.
    """
    tag = section_template["tag"]
    remove_tag = section_template.get("remove_tag", False)
    add_log2_tag = section_template.get("log2", False) and not remove_tag
    if remove_tag:
        headers = {col: col.replace(tag, "").strip() for col in columns}
    else:
        headers = {col: col for col in columns}

    if add_log2_tag:
        headers = {c: f"{h} {log2_tag}" for c, h in headers.items()}
    return headers


def eval_tag_sample_supheader(
    section_template: dict,
    log2_tag: str,
) -> str:
    """Returns the supheader name for a tag sample section.

    Args:
        columns: A list of column names to select from.
        section_template: A dictionary with "tag" containing the substring that will be
            removed from the headers if "remove_tag" is True. The "log2" key determines
            whether to add the `log2_tag` to the headers, however, if "remove_tag" is
            True the `log2_tag` will never be added. The "remove_tag" and "log"
            keys are optional and by default False.
        log2_tag: The substring that will be added to the column names if `log2` is
            True.

    Returns:
        The supheader name for the section.
    """
    supheader = section_template.get("supheader", section_template["tag"])
    if section_template.get("log2", False):
        supheader = f"{supheader} {log2_tag}"
    return supheader


def eval_column_formats(
    columns: Sequence[str],
    section_template: dict,
    format_templates: dict,
    default_format: Optional[dict] = None,
) -> dict[str, dict]:
    """Returns format descriptions for each column in the section.

    If "border" is set to True in the `section_template`, the format descriptions for
    the first and last column are updated to include borders.

    Args:
        columns: A list of column names.
        section_template: A dictionary containing the format names for columns.
        format_templates: A dictionary containing the format descriptions for each
            format name.
        default_format: Optional, the format description to use if no general format and
            no column format is specified in the `format_templates`.

    Returns:
        A dictionary containing format descriptions for each column.
    """
    if not columns:
        return {}
    default_format = {} if default_format is None else default_format
    section_format = section_template.get("format", None)
    column_formats = {}
    for col in columns:
        format_name = section_format
        if "column_format" in section_template:
            format_name = section_template["column_format"].get(col, section_format)

        column_formats[col] = format_templates.get(format_name, default_format).copy()
    if section_template.get("border", False):
        column_formats[columns[0]]["left"] = BORDER_TYPE
        column_formats[columns[-1]]["right"] = BORDER_TYPE
    return column_formats


def eval_column_conditional_formats(
    columns: Iterable[str],
    section_template: dict,
    format_templates: dict,
) -> dict[str, dict]:
    """Returns conditional format descriptions for each column in the section.

    Args:
        columns: A list of column names.
        section_template: A dictionary containing the conditional format names for
            columns.
        format_templates: A dictionary containing the conditional format descriptions
            for each format name. If a format name is not present in the
            `format_templates`, an empty dictionary is used instead.

    Returns:
        A dictionary containing conditional format descriptions for each column.
    """
    default_format = {}
    column_formats = {}
    for col in columns:
        format_name = None
        if "column_conditional" in section_template:
            format_name = section_template["column_conditional"].get(col, None)
        column_formats[col] = format_templates.get(format_name, default_format).copy()
    return column_formats


def eval_column_widths(
    columns: Iterable[str],
    section_template: dict,
    default_width: float = 64,
) -> dict[str, float]:
    """Returns column widths for each column in the section.

    Args:
        columns: A list of column names.
        section_template: A dictionary containing the column widths for columns.
        default_width: The default column width to use if no column width is specified
            in the `section_template`.

    Returns:
        A dictionary containing column widths for each column.
    """
    column_widths = {}
    for col in columns:
        column_widths[col] = section_template.get("width", default_width)
    return column_widths


def eval_header_formats(
    columns: Sequence[str], section_template: dict, format_templates: dict
) -> dict[str, dict]:
    """Returns format descriptions for each column header in the section.

    Header format descriptions defined in the `section_template` update the one from the
    `format_templates`. If "border" is set to True in the `section_template`, the
    header format descriptions for the first and last column are updated to include
    borders.

    Args:
        columns: A list of column names.
        section_template: A dictionary that can contain a "header_format" description.
        format_templates: A dictionary that can contain a "header" format description.

    Returns:
        A dictionary containing header format descriptions for each column.
    """
    if not columns:
        return {}
    temmplate_format = format_templates.get("header", {})
    section_format = section_template.get("header_format", {})
    header_format = dict(temmplate_format, **section_format)
    column_header_formats = {col: header_format.copy() for col in columns}
    if section_template.get("border", False):
        column_header_formats[columns[0]]["left"] = BORDER_TYPE
        column_header_formats[columns[-1]]["right"] = BORDER_TYPE
    return column_header_formats


def eval_supheader_format(section_template: dict, format_templates: dict) -> dict:
    """Returns a format descriptions for the supheader.

    Supheader format description defined in the `section_template` updates the one from
    the `format_templates`. If "border" is set to True in the `section_template`, the
    supheader format description is updated to include borders.

    Args:
        columns: A list of column names.
        section_template: A dictionary that can contain a "supheader_format"
            description.
        format_templates: A dictionary that can contain a "supheader" format description.

    Returns:
        A dictionary describing the supheader format.
    """
    temmplate_format = format_templates.get("supheader", {})
    section_format = section_template.get("supheader_format", {})
    supheader_format = dict(temmplate_format, **section_format)
    if section_template.get("border", False):
        supheader_format.update({"left": BORDER_TYPE, "right": BORDER_TYPE})

    return supheader_format


def eval_section_conditional_format(
    section_template: dict, format_templates: dict
) -> dict:
    """Returns a conditional format description of a section.

    Args:
        section_template: A dictionary that can contain a conditional format name with
            the key "conditional".
        format_templates: A dictionary containing the conditional format descriptions
            for each conditional format name. If a format name is not present in the
            `format_templates`, an empty dictionary is used instead.

    Returns:
        A dictionary containing a conditional format description.
    """
    section_format_name = section_template.get("conditional", None)
    section_conditional = format_templates.get(section_format_name, {}).copy()
    return section_conditional


def identify_template_section_category(section_template: dict) -> SectionCategory:
    """Identify the category of a section template.

    Args:
        section_template: A dictionary containing the section template.

    Returns:
        A SectionCategory enum value.
    """
    is_comp_group = section_template.get("comparison_group", False)
    has_tag = "tag" in section_template
    has_columns = "columns" in section_template

    if is_comp_group and has_columns and has_tag:
        return SectionCategory.COMPARISON
    if has_tag and not has_columns and not is_comp_group:
        return SectionCategory.TAG_SAMPLE
    if has_columns and not has_tag and not is_comp_group:
        return SectionCategory.STANDARD
    return SectionCategory.UNKNOWN


def _intensities_in_logspace(data: pd.DataFrame | np.ndarray | Iterable) -> bool:
    """Evaluates whether intensities are likely to be log transformed.

    Assumes that intensities are log transformed if all values are smaller or
    equal to 64. Intensities values (and intensity peak areas) reported by
    tandem mass spectrometery typically range from 10^3 to 10^12. To reach log2
    transformed values greather than 64, intensities would need to be higher
    than 10^19, which seems to be very unlikely to be ever encountered.

    Args:
        data: Dataset that contains only intensity values, can be any iterable,
            a numpy.array or a pandas.DataFrame, multiple dimensions or columns
            are allowed.

    Returns:
        True if intensity values in 'data' appear to be log transformed.
    """
    data = np.array(data, dtype=float)
    mask = np.isfinite(data)
    return np.all(data[mask].flatten() <= 64)



import os
import warnings

import click

import xlsxreport


@click.command()
def cli() -> None:
    """Description of the XlsxReport function."""

    data_dir = xlsxreport.locate_data_dir()
    if os.path.isdir(data_dir):
        click.echo(f"App data directory for XlsxReport found at:")
        click.echo(f"\t{data_dir}")
    else:
        click.echo(f"Creating XlsxReport folder in the local user data directory at:")
        click.echo(f"\t{data_dir}")
    click.echo("Copying default XlsxReport config files to the app data directory ...")
    xlsxreport.setup_data_dir()
    click.echo(f"\tConfig files were successfully copied.")



import os
import shutil
from typing import Union

import appdirs


def locate_data_dir():
    APPNAME = "XlsxReport"
    return appdirs.user_data_dir(appname=APPNAME, appauthor=False)


def setup_data_dir():
    """Creates a user specific app data directory and copies default config files."""
    data_dir = locate_data_dir()
    if not os.path.isdir(data_dir):
        os.makedirs(data_dir)

    config_dir = os.path.join(os.path.dirname(__file__), "default_config")
    for filename in os.listdir(config_dir):
        src_path = os.path.join(config_dir, filename)
        dest_path = os.path.join(data_dir, filename)
        shutil.copy(src_path, dest_path)


def get_config_file(filename: str) -> Union[str, None]:
    """Returns the file path if filename is present in the app data directory."""
    file_path = None
    data_dir = locate_data_dir()

    for subdir, dirs, files in os.walk(data_dir):
        for file in files:
            if file == filename:
                file_path = os.path.join(subdir, file)
    return file_path



import pytest
import os
from xlsxreport.template import ReportTemplate


@pytest.fixture()
def temp_template_path(request, tmp_path):
    output_path = os.path.join(tmp_path, "template_save.yaml")

    def teardown():
        if os.path.isfile(output_path):
            os.remove(output_path)

    request.addfinalizer(teardown)
    return output_path


@pytest.fixture()
def temp_template_file(request, tmp_path):
    template_path = os.path.join(tmp_path, "template_load.yaml")

    def teardown():
        if os.path.isfile(template_path):
            os.remove(template_path)

    template_text = """
        %YAML 1.2
        ---
        groups:
            group_1: {
                format: "str",
                width: 70,
                columns: ["Column 1", "Column 2",],
                column_format: {"Column 1": "int",},
                column_conditional: {"Column 1": "conditional",},
            }
            group_2: {
                columns: ["Column 3"],
            }

        formats:
            int: {"align": "center", "num_format": "0"}
            str: {"align": "left", "num_format": "0"}
            header: {"bold": True, "align": "center",}

        conditional_formats:
            conditional: {
                "type": "2_color_scale", "min_color": "#ffffbf", "max_color": "#f25540"
            }

        args:
            header_height: 95
            column_width: 45
    """
    template_text = "\n".join([line[8:] for line in template_text.splitlines()]).strip()
    with open(template_path, "w", encoding="utf-8") as file:
        file.write(template_text)
    request.addfinalizer(teardown)
    return template_path


class TestReportTemplate:
    def test_config_identical_after_load_save_reload(
        self, temp_template_file, temp_template_path
    ):
        template = ReportTemplate.load(temp_template_file)
        template.save(temp_template_path)
        loaded_template = ReportTemplate.load(temp_template_file)
        assert template.sections == loaded_template.sections
        assert template.formats == loaded_template.formats
        assert template.conditional_formats == loaded_template.conditional_formats
        assert template.settings == loaded_template.settings



"""This module provides a class for writing compiled TableSections to an Excel file."""
from __future__ import annotations
from typing import Collection, Iterable, Optional

import xlsxwriter.format
import xlsxwriter.worksheet

from xlsxreport.compiler import TableSection


class TableSectionWriter:
    """Class for writing compiled `TableSections` to an Excel file."""

    def __init__(self, workbook: xlsxwriter.Workbook):
        self.workbook = workbook
        self._xlsxwriter_formats: dict = {}  # use dictionary hash as key

    def write_sections(
        self,
        worksheet: xlsxwriter.worksheet.Worksheet,
        sections: Iterable[TableSection],
        settings: Optional[dict] = None,
        start_row: int = 0,
        start_column: int = 0,
    ) -> None:
        """Write a list of sections to the workbook to create a table.

        Args:
            worksheet: The Excel worksheet to write to.
            sections: A list of compiled `TableSection`s that will be written as a
                continuous table to the `worksheet`.
            settings: Optional, specify general settings for the table. The following
                settings are available: `write_supheader`, `supheader_height`, and
                `header_height`.
            start_row: The row in the Excel worksheet to start writing the sections at.
                The first row as seen in Excel starts at 0. The default is 0.
            start_column: The column in the Excel worksheet to start writing the
                sections at. The first column as seen in Excel starts at 0. The default
                is 0.
        """
        settings: dict = settings if settings is not None else {}
        write_supheader: bool = settings.get("write_supheader", True)
        supheader_height: float = settings.get("supheader_height", 20)
        header_height: float = settings.get("header_height", 20)
        add_autofiler: bool = settings.get("add_autofilter", True)
        freeze_cols: int = settings.get("freeze_cols", 1)

        header_row = start_row
        values_row = start_row + 1
        if write_supheader:
            header_row += 1
            values_row += 1
        next_column = start_column
        last_value_row = start_row

        for section in sections:
            self._write_section(
                worksheet=worksheet,
                section=section,
                start_row=start_row,
                start_column=next_column,
                write_supheader=write_supheader,
            )
            last_value_row = max(last_value_row, section.data.shape[0] + header_row)
            next_column += section.data.shape[1]

        # TODO - not tested from here on (including calculation of last_value_row)
        if write_supheader:
            worksheet.set_row_pixels(start_row, supheader_height)
        worksheet.set_row_pixels(header_row, header_height)
        if freeze_cols > 0:
            worksheet.freeze_panes(values_row, start_column + freeze_cols)
        if add_autofiler:
            worksheet.autofilter(
                header_row,
                start_column,
                last_value_row,
                last_col=next_column - 1,
            )

    def _write_section(
        self,
        worksheet: xlsxwriter.worksheet.Worksheet,
        section: TableSection,
        start_row: int,
        start_column: int,
        write_supheader: bool,
    ) -> None:
        """Write a TableSection to the workbook."""
        num_values, num_rows = section.data.shape
        header_row = start_row
        values_row = start_row + 1

        if write_supheader:
            header_row += 1
            values_row += 1
            self._write_supheader(
                worksheet=worksheet,
                row=start_row,
                column=start_column,
                num_columns=section.data.shape[1],
                supheader=section.supheader,
                supheader_format=section.supheader_format,
            )
        for column_position, column in enumerate(section.data.columns):
            self._write_column(
                worksheet=worksheet,
                row=header_row,
                column=start_column + column_position,
                header=section.headers[column],
                values=section.data[column],
                header_format=section.header_formats[column],
                values_format=section.column_formats[column],
                conditional_format=section.column_conditionals[column],
                column_width=section.column_widths[column],
            )
        if section.section_conditional:
            worksheet.conditional_format(
                values_row,
                start_column,
                values_row + num_values - 1,
                start_column + num_rows - 1,
                section.section_conditional,
            )
        if section.hide_section:
            worksheet.set_column(
                start_column,
                start_column + num_rows - 1,
                options={"level": 1, "collapsed": True, "hidden": True},
            )

    def _write_supheader(
        self,
        worksheet: xlsxwriter.worksheet.Worksheet,
        row: int,
        column: int,
        num_columns: int,
        supheader: str,
        supheader_format: dict[str, float | str | bool],
    ) -> None:
        """Write a supheader to the workbook by merging a range of cells."""
        supheader_xlsx_format = self.get_xlsx_format(supheader_format)
        if not supheader:
            return
        if num_columns > 1:
            last_column = column + num_columns - 1
            worksheet.merge_range(
                row, column, row, last_column, supheader, supheader_xlsx_format
            )
        else:
            worksheet.write(row, column, supheader, supheader_xlsx_format)

    def _write_column(
        self,
        worksheet: xlsxwriter.worksheet.Worksheet,
        row: int,
        column: int,
        header: str,
        values: Collection,
        header_format: dict[str, float | str | bool],
        values_format: dict[str, float | str | bool],
        conditional_format: dict[str, float | str | bool],
        column_width: float,
    ) -> None:
        """Write a column to the workbook."""
        header_xlsx_format = self.get_xlsx_format(header_format)
        values_xlsx_format = self.get_xlsx_format(values_format)
        worksheet.write(row, column, header, header_xlsx_format)
        worksheet.write_column(row + 1, column, values, values_xlsx_format)
        worksheet.set_column_pixels(column, column, column_width)
        if conditional_format:
            worksheet.conditional_format(
                row + 1, column, row + len(values), column, conditional_format
            )

    def get_xlsx_format(
        self, format_description: dict[str, float | str | bool]
    ) -> xlsxwriter.format.Format:
        """Converts a format description to an `xlsxwriter` format.

        Args:
            format_description: A dictionary describing the format.
        """
        _hash = _hashable_from_dict(format_description)
        if _hash not in self._xlsxwriter_formats:
            self._xlsxwriter_formats[_hash] = self.workbook.add_format(
                format_description
            )
        return self._xlsxwriter_formats[_hash]


def _hashable_from_dict(format_description: dict[str, float | str | bool]):
    return tuple((k, format_description[k]) for k in sorted(format_description))



from io import BytesIO
from unittest.mock import MagicMock, call

import openpyxl
import pandas as pd
import pytest
from xlsxwriter import Workbook as Workbook

import xlsxreport.compiler as compiler
import xlsxreport.excel_writer as writer


class ExcelWriteReadTestManager:
    """Helper class for writing an excel file with `xlsxwriter` to buffer and reading it
    from the buffer with `openpyxl`.

    Note that for writing rows and columns xlsxwriter is zero indexed whereas pyopenxl
    is one indexed.
    """

    def __init__(self):
        self.buffer = None
        self.loaded_workbook = None
        self.loaded_worksheet = None

    def __enter__(self):
        self.buffer = BytesIO()
        self.workbook = Workbook(self.buffer)
        self.worksheet_name = "Worksheet"
        self.worksheet = self.workbook.add_worksheet(self.worksheet_name)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.workbook.close()
        self._load_excel_from_buffer()

    def _load_excel_from_buffer(self):
        if self.buffer is None:
            raise Exception("Buffer is not initialized")
        self.loaded_workbook = openpyxl.load_workbook(self.buffer)
        self.loaded_worksheet = self.loaded_workbook[self.worksheet_name]


@pytest.fixture()
def table_section() -> compiler.TableSection:
    section = compiler.TableSection(
        data=pd.DataFrame({"Column 1": [1, 2, 3], "Column 2": ["A", "B", "C"]}),
        column_widths={"Column 1": 10, "Column 2": 10},
    )
    return section


class TestTableSectionWriteSections:
    """Tests for TableSectionWriter.write_section():
    - Test that multiple sections are written with correct column and row positions
    - Test that supheader, header and row heights are set with correct values and on
    correct coordinates
    - Test that freeze panes is applied to the correct coordinates
    - Test that auto filter is applied to the correct coordinates
    """

    @pytest.fixture(autouse=True)
    def _init_table_sections(self):
        section_1 = compiler.TableSection(
            data=pd.DataFrame({"Column 1": [1, 2, 3], "Column 2": ["A", "B", "C"]}),
        )
        section_2 = compiler.TableSection(
            data=pd.DataFrame({"Column 3": [1, 2, 3], "Column 4": ["A", "B", "C"]}),
        )
        self.headers = list(section_1.headers) + list(section_2.headers)
        self.table_sections = [section_1, section_2]
        self.col_num = 4
        self.row_num = 3
        self.column_values = []
        for section in self.table_sections:
            self.column_values.extend([section.data[c].tolist() for c in section.data.columns])  # fmt: skip

    @pytest.mark.parametrize("start_row", [0, 2, 10])
    def test_all_section_headers_correctly_written_with_different_start_rows(self, start_row):  # fmt: skip
        with ExcelWriteReadTestManager() as excel_manager:
            section_writer = writer.TableSectionWriter(excel_manager.workbook)
            section_writer.write_sections(
                excel_manager.worksheet,
                self.table_sections,
                settings={"write_supheader": False},
                start_row=start_row,
            )
        sheet = excel_manager.loaded_worksheet
        assert [cell.value for cell in list(sheet.rows)[start_row]] == self.headers

    def test_all_section_values_are_written_to_the_correct_position(self):
        with ExcelWriteReadTestManager() as excel_manager:
            section_writer = writer.TableSectionWriter(excel_manager.workbook)
            section_writer.write_sections(
                excel_manager.worksheet,
                self.table_sections,
                settings={"write_supheader": False},
            )
        sheet_columns = list(excel_manager.loaded_worksheet.columns)
        for column_cells, column_values in zip(sheet_columns, self.column_values):
            value_cells = column_cells[1:]  # without header cell
            assert [cell.value for cell in value_cells] == column_values

    @pytest.mark.parametrize("start_col", [0, 2, 10])
    def test_start_column_correctly_applied(self, start_col):
        with ExcelWriteReadTestManager() as excel_manager:
            section_writer = writer.TableSectionWriter(excel_manager.workbook)
            section_writer.write_sections(
                excel_manager.worksheet, self.table_sections, start_column=start_col
            )
        sheet = excel_manager.loaded_worksheet
        for empty_column in list(sheet.columns)[:start_col]:
            assert all([cell.value is None for cell in empty_column])
        assert len(list(sheet.columns)) == self.col_num + start_col

    @pytest.mark.parametrize(
        "write_supheader, start_row", [(True, 0), (True, 2), (False, 0), (False, 2)]
    )
    def test_start_row_and_supheader_correctly_applied(
        self, write_supheader, start_row
    ):
        with ExcelWriteReadTestManager() as excel_manager:
            section_writer = writer.TableSectionWriter(excel_manager.workbook)
            section_writer.write_sections(
                excel_manager.worksheet,
                self.table_sections,
                settings={"write_supheader": write_supheader},
                start_row=start_row,
            )
        sheet_rows = list(excel_manager.loaded_worksheet.rows)
        header_row = start_row + int(write_supheader)
        assert [cell.value for cell in sheet_rows[header_row]] == self.headers


class TestIntegrationTableSectionWriteColumn:
    @pytest.fixture(autouse=True)
    def _init_section_writer_with_write_column(self):
        # Note that xlsxwriter is zero indexed whereas pyopenxl is one indexed
        row, column = 2, 2
        with ExcelWriteReadTestManager() as excel_manager:
            section_writer = writer.TableSectionWriter(excel_manager.workbook)
            section_writer._write_column(
                excel_manager.worksheet,
                row=row - 1,
                column=column - 1,
                header="Header",
                values=[1, 2, "3"],
                header_format={"bold": True, "bottom": 2},
                values_format={"align": "center", "num_format": "0.00"},
                conditional_format={"type": "2_color_scale"},
                column_width=10000,
            )
        self.worksheet = excel_manager.loaded_worksheet
        self.written_column = list(self.worksheet.columns)[1]

    def test_column_is_written_to_correct_position(self):
        columns = list(self.worksheet.columns)
        empty_col_position = 0
        written_col_position = 1
        assert all([cell.value is None for cell in columns[empty_col_position]])
        assert any([cell.value is None for cell in columns[written_col_position]])

    def test_correct_column_values_written(self):
        column_values = [cell.value for cell in self.written_column]
        assert column_values == [None, "Header", 1, 2, "3"]

    def test_header_format_is_applied(self):
        header_cell = self.worksheet.cell(row=2, column=2)
        assert header_cell.font.bold == True
        assert header_cell.border.bottom.style == "medium"

    def test_column_format_is_applied(self):
        column_cells = self.written_column[2:]  # without empty and header cell
        assert all([cell.alignment.horizontal == "center" for cell in column_cells])
        assert all([cell.number_format == "0.00" for cell in column_cells])

    def test_set_correct_column_width(self):
        # xlsxwriter sets column width in pixel units, whereas openpyxl sets width in
        # some kind of "unit" format. So we set the width to a high number and check if
        # the column width in "units" is smaller or bigger than 100.
        empty_col_width = self.worksheet.column_dimensions["A"].width
        written_col_width = self.worksheet.column_dimensions["B"].width
        assert empty_col_width < 100
        assert written_col_width > 100

    def test_conditional_format_applied_to_correct_area(self):
        conditional_format = list(self.worksheet.conditional_formatting)[0]
        conditional_format.cells.ranges[0].top == [(3, 2)]
        conditional_format.cells.ranges[0].bottom == [(5, 2)]


class TestTableSectionWriteColumn:
    @pytest.fixture(autouse=True)
    def _init_section_writer(self):
        self.worksheet_mock = MagicMock(name="worksheet_mock")
        self.section_writer = writer.TableSectionWriter(Workbook())

    @pytest.fixture(autouse=True)
    def _init_arguments_for_write_column(self):
        self.args = {
            "row": 1,
            "column": 1,
            "header": "Header name",
            "values": [1, 2, 3],
            "header_format": {"bold": True},
            "values_format": {"bold": False},
            "conditional_format": {"type": "cell"},
            "column_width": 10,
        }
        self.header_row = self.args["row"]
        self.data_row_start = self.args["row"] + 1
        self.data_row_end = self.args["row"] + len(self.args["values"])

    def test_write_column_called_with_correct_arguments_to_write_values(self):
        self.section_writer._write_column(self.worksheet_mock, **self.args)
        assert self.worksheet_mock.write_column.call_args == call(
            self.data_row_start,
            self.args["column"],
            self.args["values"],
            self.section_writer.get_xlsx_format(self.args["values_format"]),
        )

    def test_write_called_with_correct_arguments_to_write_header(self):
        self.section_writer._write_column(self.worksheet_mock, **self.args)
        assert self.worksheet_mock.write.call_args == call(
            self.header_row,
            self.args["column"],
            self.args["header"],
            self.section_writer.get_xlsx_format(self.args["header_format"]),
        )

    def test_set_column_pixels_called_with_correct_arguments(self):
        self.section_writer._write_column(self.worksheet_mock, **self.args)
        assert self.worksheet_mock.set_column_pixels.call_args == call(
            self.args["column"], self.args["column"], self.args["column_width"]
        )

    def test_conditional_format_called_with_correct_arguments(self):
        self.section_writer._write_column(self.worksheet_mock, **self.args)

        assert self.worksheet_mock.conditional_format.call_args == call(
            self.data_row_start,
            self.args["column"],
            self.data_row_end,
            self.args["column"],
            self.args["conditional_format"],
        )

    def test_conditional_format_not_called_when_conditionaL_format_is_empty(self):
        self.args["conditional_format"] = {}
        self.section_writer._write_column(self.worksheet_mock, **self.args)
        self.worksheet_mock.conditional_format.assert_not_called()


class TestIntegrationTableSectionWriteSupheader:
    def _create_worksheet_with_section_writer_and_write_supheader(
        self, row: int = 1, column: int = 1, num_columns: int = 1, supheader = "Supheader"  # fmt: skip
    ):
        """Creates an xlsxwriter.workbook, an xlsxwriter.worksheet and a
        TableSectionWriter. After using the TableSectionWriter to write a supheader to
        the worksheet, it is then safed to a buffer and loaded with openpyxl. The loaded
        worksheet is returned.

        Note that row and column are specified as one indexed, whereas xlsxwriter is
        zero indexed.
        """
        with ExcelWriteReadTestManager() as excel_manager:
            section_writer = writer.TableSectionWriter(excel_manager.workbook)
            section_writer._write_supheader(
                worksheet=excel_manager.worksheet,
                row=row - 1,
                column=column - 1,
                num_columns=num_columns,
                supheader=supheader,
                supheader_format={"bold": True},
            )
        return excel_manager.loaded_worksheet

    @pytest.mark.parametrize("row, column", [(1, 2), (2, 1), (5, 10)])
    def test_supheader_written_to_correct_position(self, row, column):
        worksheet = self._create_worksheet_with_section_writer_and_write_supheader(row, column)  # fmt: skip
        assert worksheet.cell(row=row, column=column).value == "Supheader"

    def test_supheader_format_is_applied(self):
        worksheet = self._create_worksheet_with_section_writer_and_write_supheader()
        assert worksheet.cell(row=1, column=1).font.bold == True

    @pytest.mark.parametrize("num_columns", [2, 4, 10])
    def test_supheader_cell_is_merged(self, num_columns):
        worksheet = self._create_worksheet_with_section_writer_and_write_supheader(num_columns=num_columns)  # fmt: skip
        merged_cells = list(worksheet.merged_cells.ranges[0].cells)
        assert merged_cells == [(1, i) for i in range(1, num_columns + 1)]

    def test_no_merge_applied_when_written_with_only_one_column(self):
        worksheet = self._create_worksheet_with_section_writer_and_write_supheader(num_columns=1)  # fmt: skip
        assert len(worksheet.merged_cells.ranges) == 0
        assert worksheet.cell(row=1, column=1).value == "Supheader"

    def test_no_merge_applied_when_no_supheader_specified(self):
        worksheet = self._create_worksheet_with_section_writer_and_write_supheader(num_columns=2, supheader="")  # fmt: skip
        assert len(worksheet.merged_cells.ranges) == 0


class TestIntegrationTableSectionWriteSection:
    def _create_worksheet_with_section_writer_and_write_section(
        self, table_section, write_supheader
    ):
        """Creates an xlsxwriter.workbook, an xlsxwriter.worksheet and a
        TableSectionWriter. After using the TableSectionWriter to write a section to the
        worksheet, it is then safed to a buffer and loaded with openpyxl. The loaded
        worksheet is returned.

        Note that row and column are specified as one indexed, whereas xlsxwriter is
        zero indexed.
        """
        with ExcelWriteReadTestManager() as excel_manager:
            section_writer = writer.TableSectionWriter(excel_manager.workbook)
            section_writer._write_section(
                excel_manager.worksheet,
                table_section,
                start_row=0,
                start_column=0,
                write_supheader=write_supheader,
            )
        return excel_manager.loaded_worksheet

    @pytest.mark.parametrize("write_supheader", [True, False])
    def test_correct_number_of_columns_written(self, table_section, write_supheader):
        worksheet = self._create_worksheet_with_section_writer_and_write_section(
            table_section, write_supheader=write_supheader
        )
        assert len(list(worksheet.columns)) == table_section.data.shape[1]

    @pytest.mark.parametrize("write_supheader, num_headers", [(True, 2), (False, 1)])
    def test_correct_number_of_rows_written(self, table_section, write_supheader, num_headers):  # fmt: skip
        worksheet = self._create_worksheet_with_section_writer_and_write_section(
            table_section, write_supheader=write_supheader
        )
        for column in worksheet.columns:
            assert len(column) == table_section.data.shape[0] + num_headers

    @pytest.mark.parametrize("write_supheader", [True, False])
    def test_supheader_correctly_added_or_ommitted(self, table_section, write_supheader):  # fmt: skip
        table_section.supheader = "A supheader name"
        worksheet = self._create_worksheet_with_section_writer_and_write_section(
            table_section, write_supheader=write_supheader
        )
        # Note that xlsxwriter is zero indexed whereas pyopenxl is one indexed.
        if write_supheader:
            assert worksheet.cell(row=1, column=1).value == table_section.supheader
        else:
            assert worksheet.cell(row=1, column=1).value != table_section.supheader


class TestTableSectionWriteSection:
    @pytest.fixture(autouse=True)
    def _init_section_writer(self):
        self.worksheet_mock = MagicMock(name="worksheet_mock")
        self.section_writer = writer.TableSectionWriter(Workbook())

    @pytest.mark.parametrize("write_supheader", [True, False])
    def test_conditional_format_called_with_correct_arguments(self, table_section, write_supheader):  # fmt: skip
        table_section.section_conditional = {"bold": True}

        self.section_writer._write_section(
            self.worksheet_mock,
            table_section,
            start_row=0,
            start_column=0,
            write_supheader=write_supheader,
        )

        rows, cols = table_section.data.shape
        data_start_row = 2 if write_supheader else 1
        data_end_row = data_start_row + rows - 1
        self.worksheet_mock.conditional_format.assert_called_once_with(
            data_start_row, 0, data_end_row, cols - 1, {"bold": True}
        )

    def test_conditional_format_not_called_when_conditionaL_format_is_empty(self, table_section):  # fmt: skip
        table_section.section_conditional = {}
        self.section_writer._write_section(self.worksheet_mock, table_section, 0, 0, True)  # fmt: skip
        self.worksheet_mock.conditional_format.assert_not_called()

    def test_set_column_not_called_when_hide_section_is_false(self, table_section):
        table_section.hide_section = False
        self.section_writer._write_section(self.worksheet_mock, table_section, 0, 0, True)  # fmt: skip
        self.worksheet_mock.set_column.assert_not_called()

    def test_set_column_called_correctly_when_hide_section_is_false(self, table_section):  # fmt: skip
        table_section.hide_section = True
        self.section_writer._write_section(self.worksheet_mock, table_section, 0, 0, True)  # fmt: skip
        self.worksheet_mock.set_column.assert_called_once_with(
            0, table_section.data.shape[1] - 1, options={"level": 1, "collapsed": True, "hidden": True}  # fmt: skip
        )


class TestTableSectionWriterGetXlsxFormat:
    @pytest.fixture(autouse=True)
    def _init_section_writer(self):
        self.writer = writer.TableSectionWriter(Workbook())
        self.format_description = {"bold": True, "font_color": "#FF0000"}

    def test_xlsx_format_with_correct_properties_is_returned(self):
        xlsx_format = self.writer.get_xlsx_format(self.format_description)
        assert xlsx_format.bold == True and xlsx_format.font_color == "#FF0000"

    def test_multiple_calls_return_same_object(self):
        xlsx_format_1 = self.writer.get_xlsx_format(self.format_description)
        xlsx_format_2 = self.writer.get_xlsx_format(self.format_description)
        assert xlsx_format_1 is xlsx_format_2


class TestHashableFromDictionary:
    @pytest.mark.parametrize(
        "dictionary", [{"A": 1, "C": 2, "B": 1}, {"B": 1, "A": 1, "C": 2}]
    )
    def test_different_key_order_creates_same_hashable(self, dictionary):
        expected_hash = writer._hashable_from_dict({"A": 1, "B": 1, "C": 2})
        assert writer._hashable_from_dict(dictionary) == expected_hash

    def test_different_dict_values_create_different_hashable(self):
        hashable_1 = writer._hashable_from_dict({"A": 1})
        hashable_2 = writer._hashable_from_dict({"A": 2})
        assert hashable_1 != hashable_2



from __future__ import annotations
from typing import Optional

import pandas as pd
import xlsxwriter

from xlsxreport.excel_writer import TableSectionWriter
from xlsxreport.template import ReportTemplate
from xlsxreport.compiler import prepare_table_sections


class Reportbook(xlsxwriter.Workbook):
    """Subclass of the XlsxWriter Workbook class."""

    def add_infosheet(self) -> xlsxwriter.worksheet.Worksheet:
        worksheet = self.add_worksheet("Info")
        return worksheet

    def add_datasheet(self, name: Optional[str] = None) -> Datasheet:
        worksheet = self.add_worksheet(name)
        data_sheet = Datasheet(self, worksheet)
        return data_sheet


class Datasheet:
    def __init__(
        self, workbook: xlsxwriter.Workbook, worksheet: xlsxwriter.worksheet.Worksheet
    ):
        self.workbook = workbook
        self.worksheet = worksheet
        self.section_writer = TableSectionWriter(self.workbook)
        self.table = None
        self.report_template = None

    def apply_configuration(self, config_file: str) -> None:
        """Reads a config file and prepares workbook formats."""
        self.report_template = ReportTemplate.load(config_file)
        self.report_template.settings["evaluate_log2_transformation"] = True

    def add_data(self, table: pd.DataFrame) -> None:
        self.table = table

    def write_data(self) -> None:
        table_sections = prepare_table_sections(self.report_template, self.table)
        self.section_writer.write_sections(
            self.worksheet, table_sections, settings=self.report_template.settings
        )
